Imports System.Math
Imports System.text

Module MultiPrecision
    Public Const RX As Integer = 256
    Public Const RXb As Integer = 8

    Const RXm As Integer = RX - 1

    Private ip() As Integer
    Private ww() As Double

    Public Sub MpADD(ByVal u() As Byte, ByVal pu As Integer, _
                     ByVal v() As Byte, ByVal pv As Integer, ByVal m As Integer, _
                     ByVal w() As Byte)
        pu += m - 1
        pv += m - 1

        Dim a(m) As Byte
        Dim pa As Integer = m

        Dim ireg As Integer = 0
        Do
            ireg = HiByte(ireg) + u(pu) + v(pv)
            a(pa) = CByte(LoByte(ireg))
            pu -= 1
            pv -= 1
            pa -= 1
        Loop Until pa = 0
        a(0) = CByte(HiByte(ireg))

        a.CopyTo(w, 0)
        Erase a
    End Sub

    Public Sub MpSUB(ByVal u() As Byte, ByVal pu As Integer, _
                     ByVal v() As Byte, ByVal pv As Integer, ByVal m As Integer, _
                     ByVal w() As Byte, ByRef isign As Integer)
        pu += m - 1
        pv += m - 1

        Dim a(m - 1) As Byte
        Dim pa As Integer = m - 1

        Dim ireg As Integer = RX
        Do
            ireg = HiByte(ireg) + RXm + u(pu) - v(pv)
            a(pa) = CByte(LoByte(ireg))
            pu -= 1
            pv -= 1
            pa -= 1
        Loop Until pa < 0
        isign = HiByte(ireg) - 1

        a.CopyTo(w, 0)
        Erase a
    End Sub

    Public Sub MpNEG(ByVal u() As Byte, ByVal pu As Integer, ByVal m As Integer)
        Dim ou As Integer = pu
        pu += m - 1

        Dim ireg As Integer = RX
        Do
            ireg = 255 - u(pu) + HiByte(ireg)
            u(pu) = CByte(LoByte(ireg))
            pu -= 1
        Loop Until pu < ou
    End Sub

    Public Sub MpSAD(ByVal u() As Byte, ByVal pu As Integer, ByVal m As Integer, _
                     ByVal iv As Integer, ByVal w() As Byte)
        pu += m - 1

        Dim a(m) As Byte
        Dim pa As Integer = m

        Dim ireg As Integer = iv << 8
        Do
            ireg = u(pu) + HiByte(ireg)
            a(pa) = CByte(LoByte(ireg))
            pu -= 1
            pa -= 1
        Loop Until pa = 0
        a(0) = CByte(HiByte(ireg))

        a.CopyTo(w, 0)
        Erase a
    End Sub

    Public Sub MpSMU(ByVal u() As Byte, ByVal pu As Integer, ByVal m As Integer, _
                     ByVal iv As Integer, ByVal w() As Byte)
        pu += m - 1

        Dim a(m) As Byte
        Dim pa As Integer = m

        Dim ireg As Integer = 0
        Do
            ireg = u(pu) * iv + HiByte(ireg)
            a(pa) = CByte(LoByte(ireg))
            pu -= 1
            pa -= 1
        Loop Until pa = 0
        a(0) = CByte(HiByte(ireg))

        a.CopyTo(w, 0)
        Erase a
    End Sub

    Public Sub MpSDV(ByVal u() As Byte, ByVal pu As Integer, ByVal m As Integer, _
                     ByVal iv As Integer, ByVal q() As Byte, ByRef ir As Integer)
        Dim a(m - 1) As Byte
        Dim pa As Integer = 0

        ir = 0
        Do
            Dim i As Integer = RX * ir + u(pu)
            a(pa) = CByte(i \ iv)
            ir = i Mod iv

            pu += 1
            pa += 1
        Loop Until pa = m

        a.CopyTo(q, 0)
        Erase a
    End Sub

    Public Sub MpMUL(ByVal u() As Byte, ByVal pu As Integer, ByVal m As Integer, _
                 ByVal v() As Byte, ByVal pv As Integer, ByVal n As Integer, _
                 ByVal w() As Byte)
        Static nnold As Integer

        Dim n_max As Integer = Max(m, n)
        Dim nm As Integer = n + m - 1

        Dim nn As Integer = 1
        Do While (nn < n_max)
            nn <<= 1
        Loop
        nn <<= 1

        If nn <> nnold Then
            ReDim ip(2 + CInt(Sqrt(nn \ 2)))
            ReDim ww(nn \ 2 - 1)
            nnold = nn
        End If

        Dim a(nn - 1) As Double
        For J As Integer = 0 To m - 1
            a(J) = u(pu + J)
        Next

        Dim b(nn - 1) As Double
        For J As Integer = 0 To n - 1
            b(j) = v(pv + j)
        Next

        rdft(a, FFT_FORWARD)
        rdft(b, FFT_FORWARD)

        b(0) *= a(0)
        b(1) *= a(1)
        For re As Integer = 2 To nn - 1 Step 2
            Dim bre As Double = b(re)
            Dim im As Integer = re + 1
            b(re) = bre * a(re) - b(im) * a(im)
            b(im) = bre * a(im) + b(im) * a(re)
        Next

        rdft(b, FFT_REVERSE)

        Dim cy As Long = 0
        For j As Integer = nn - 1 To 0 Step -1
            Dim t As Long = CLng(b(j) / (nn >> 1)) + cy
            cy = t >> RXb
            If j < nm Then w(j + 1) = CByte(t And RXm)
        Next

        If cy > RXm Then Throw New System.ArithmeticException("cannot happen in fftmul")
        w(0) = CByte(cy)

        Erase a, b
    End Sub

    Public Sub MpINV(ByVal u() As Byte, ByVal pu As Integer, ByVal m As Integer, ByVal v() As Byte)
        Const MF As Integer = 8
        Const BI As Double = 1 / RX
        Const RXm2 As Byte = RX - 2

        Dim n As Integer = v.Length
        Dim mm As Integer = Min(MF, m)

        '================================================================
        'Use ordinary floating arithmetic to get an initial approximation
        '================================================================
        Dim fu As Double = u(pu + mm - 1)
        For j As Integer = mm - 2 To 0 Step -1
            fu *= BI
            fu += u(pu + j)
        Next

        Dim fv As Double = 1 / fu

        Dim v1(2 * n - 1) As Byte
        Dim rpv1 As Integer = 0

        For j As Integer = 0 To n - 1
            Dim i As Double = Floor(fv)
            v1(j) = CByte(i)
            fv = RX * (fv - i)
        Next

        '=====================================
        'Iterate Newton's rule to convergence.
        '=====================================
        Dim s(n + m - 1) As Byte
        Do
            '---------------------
            'Construct 2 - VU in S
            '---------------------
            's = v1 * u
            's : n + m digits with the radix point after s(1)
            MpMUL(v1, rpv1, n, u, pu, m, s)

            'calculate 2 - s with n digits
            'the radix point is located after s(1)
            MpNEG(s, 1, n)
            s(1) -= RXm2

            '------------------
            'Multiply SV into V
            '------------------
            'v1 = s * v1
            'v1 : n + n digits with the radix point after v1(1)
            MpMUL(s, 1, n, v1, rpv1, n, v1)
            rpv1 = 1

            '--------------------------------------------------------------
            'If fractional part of S is not zero, it has not converged to 1
            '--------------------------------------------------------------
            Dim j As Integer
            For j = 2 To n - 1
                If s(j) > 0 Then Exit For
            Next

            If j = n Then Exit Do
        Loop

        '==========================================
        'store the most significant n digits into V
        '==========================================
        Array.Copy(v1, 1, v, 0, n)
        Erase s, v1
    End Sub

    Public Sub MpSQRT(ByVal u() As Byte, ByVal pu As Integer, ByVal m As Integer, _
                      ByVal v() As Byte, ByVal w() As Byte)
        Const MF As Integer = 8
        Const BI As Double = 1 / RX

        Dim n As Integer = w.Length
        Dim mm As Integer = Min(MF, m)

        Dim fu As Double = u(pu + mm - 1)
        For j As Integer = mm - 2 To 0 Step -1
            fu *= BI
            fu += u(pu + j)
        Next

        Dim fw As Double = 1 / Sqrt(fu)

        Dim w1(2 * n - 1) As Byte
        For j As Integer = 0 To n - 1
            Dim i As Double = Floor(fw)
            w1(j) = CByte(i)
            fw = RX * (fw - i)
        Next

        MySQRT(u, pu, m, v, w, w1)

        Erase w1
    End Sub

    Public Sub MpSQRT(ByVal u() As Byte, ByVal pu As Integer, ByVal m As Integer, _
                      ByVal v() As Byte, ByVal w() As Byte, ByVal wi() As Byte)
        Dim n As Integer = w.Length
        Dim w1(2 * n - 1) As Byte

        For j As Integer = 0 To n - 1
            w1(j) = wi(j)
        Next

        MySQRT(u, pu, m, v, w, w1)

        Erase w1
    End Sub

    Private Sub MySQRT(ByVal u() As Byte, ByVal pu As Integer, ByVal m As Integer, _
                       ByVal v() As Byte, ByVal w() As Byte, ByVal w1() As Byte)
        Const RXm3 As Byte = RX - 3

        Dim n As Integer = w.Length
        Dim rpw1 As Integer = 0

        '=====================================
        'Iterate Newton's rule to convergence.
        '=====================================
        Dim s(n + Max(n, m) - 1) As Byte
        Do
            '-----------------------------
            'Construct S = (3 - U W W) / 2
            '-----------------------------
            's = w * w
            's : n + n digits with the radix point after r(1)
            MpMUL(w1, rpw1, n, w1, rpw1, n, s)

            's = u * s
            '  = u * w * w
            's : n + m digits with the radix point after s(1)
            '    only n digits of s are used.
            MpMUL(s, 1, n, u, pu, Min(m, n), s)

            'calculate 3 - s with n digits
            'the radix point is located after s(1)
            MpNEG(s, 1, n)
            s(1) -= RXm3

            'divide by 2
            'the radix point is located after s(0)
            Dim ir As Integer
            MpSDV(s, 1, n, 2, s, ir)

            '--------------------------------------------------------------
            'If fractional part of S is not zero, it has not converged to 1
            '--------------------------------------------------------------
            Dim j As Integer
            For j = 1 To n - 2
                If s(j) > 0 Then Exit For
            Next

            If j = n - 1 Then Exit Do

            '--------------------
            'Replace w1 by s * w1
            '--------------------
            'w1 = s * w1
            'n digits of s and n digits of w1 produces
            'n + n digits of w1.
            'the radix point is located after t(1)
            MpMUL(s, 0, n, w1, rpw1, n, w1)
            rpw1 = 1
        Loop

        '==========================================
        'Get square root from reciprocal and return
        '==========================================
        'v = s = u * w1
        'v = 1/sqrt(u) * u = sqrt(u)
        MpMUL(w1, rpw1, n, u, pu, Min(m, n), s)

        Array.Copy(w1, rpw1, w, 0, n)
        Array.Copy(s, 1, v, 0, n)
    End Sub

    Public Sub MpDIV(ByVal u() As Byte, ByVal v() As Byte, _
                     ByRef q() As Byte, ByRef r() As Byte)
        Const MACC As Integer = 6

        Dim n As Integer = u.Length
        Dim m As Integer = v.Length

        Dim u1(n - 1) As Byte
        u.CopyTo(u1, 0)
        Dim v1(m - 1) As Byte
        v.CopyTo(v1, 0)

        '-------------
        'Set S = 1 / V
        '-------------
        Dim s(n + MACC - 1) As Byte
        MpINV(v, 0, m, s)

        '-----------
        'Set Q = S U
        '-----------
        'rr = s * u : n + n + MACC bytes
        '  s : n + MACC bytes
        '  u : n bytes
        Dim rr(2 * n + MACC - 1) As Byte
        MpMUL(s, 0, n + MACC, u, 0, n, rr)
        MpSAD(rr, 0, n + MACC - 1, 1, s)

        If (q Is Nothing) OrElse (q.Length < n - m + 1) Then ReDim q(n - m)
        Array.Copy(s, 2, q, 0, n - m + 1)

        '------------------------------------------
        'Multiply and subtract to get the remainder
        '------------------------------------------
        MpMUL(q, 0, n - m + 1, v1, 0, m, rr)
        Dim isign As Integer
        MpSUB(u1, 0, rr, 1, n, rr, isign)
        If (r Is Nothing) OrElse (r.Length < m) Then ReDim r(m - 1)
        Array.Copy(rr, n - m, r, 0, m)

        Erase s, rr, u1, v1
    End Sub

    Private Function LoByte(ByVal x As Integer) As Integer
        Return x And RXm
    End Function

    Private Function HiByte(ByVal x As Integer) As Integer
        Return (x >> RXb) And RXm
    End Function

End Module
