Imports System.Text
Imports System.Math

Module RadixConv
    Private Const BYTES_THRESHOLD As Integer = 512
    Private Const DIGITS_THRESHOLD As Integer = 55000

    Public Function RadixToDec(ByVal u() As Byte, ByVal pu As Integer, ByVal m As Integer) As String
        Dim a(m - 1) As Byte
        Array.Copy(u, pu, a, 0, m)
        Return RadixToDec(a)
    End Function

    Private Function RadixToDec(ByVal a() As Byte) As String
        '===================================================
        'If the data length is smaller than BYTES_THRESHOLD,
        'conversion should be done by a O(N2) method
        '===================================================
        If a.Length <= BYTES_THRESHOLD Then
            Dim sr As New StringBuilder(CInt(BYTES_THRESHOLD * 2.05))
            Dim I As Integer = 0
            Do
                Do
                    If a(I) <> 0 Then Exit Do
                    If I = a.Length - 1 Then
                        Return sr.ToString
                    End If
                    I += 1
                Loop

                Dim ir As Integer
                MpSDV(a, 0, a.Length, 10, a, ir)
                sr.Insert(0, ir)
            Loop
        End If

        '===========================================
        'Search a power of 10 of a half length bytes
        '===========================================
        Dim dc As Integer = CInt(2.408 * a.Length) \ 2
        Dim w() As Byte = Power10(dc)

        '=================================================
        'Get a quotient and a remainder of the power of 10
        'which are higher and lower parts
        '=================================================
        Dim ah() As Byte, al() As Byte
        MpDIV(a, w, ah, al)

        '=================================================
        'Radix conversion of both parts by calling myself
        '=================================================
        Dim s As String = RadixToDec(ah) & RadixToDec(al).PadLeft(dc, "0"c)

        Erase ah, al, a, w
        Return s
    End Function

    Public Function RadixToDecFrac(ByVal u() As Byte, ByVal pu As Integer, ByVal dc As Integer) As String
        Dim LenFrac As Integer = CInt(dc / Log10(256)) + 5

        '====================================================
        'If the data length is smaller than DIGITS_THRESHOLD,
        'conversion should be done by a O(N2) method
        '====================================================
        If dc <= DIGITS_THRESHOLD Then
            Dim sr As New StringBuilder(dc)
            Dim a(LenFrac) As Byte

            MpSMU(u, pu, LenFrac, 10, a)
            sr.Append(a(0))

            Do Until sr.Length = dc
                MpSMU(a, 1, LenFrac, 10, a)
                sr.Append(a(0))
            Loop

            Return sr.ToString
        End If

        '=================
        'calculate 10 ** n
        '=================
        Dim w() As Byte = Power10(dc)
        Dim LenP10 As Integer = w.Length

        '=======================================================
        'Multiply a by w to move a fractional part of a desired
        'digit number to an integer part
        '=======================================================
        ReDim Preserve w(LenP10 + LenFrac - 1)
        MpMUL(w, 0, LenP10, u, pu, LenFrac, w)

        '====================================
        'Radix conversion of the integer part
        '====================================
        Dim s As String = RadixToDec(w, 0, LenP10)

        Return s
    End Function

    Private Function Power10(ByVal dc As Integer) As Byte()
        Dim w() As Byte
        Dim wp As Integer
        Dim wlen As Integer

        Do
            Dim L As Integer = CInt(Floor(Log(dc) / Log(2)))
            Dim LL As Integer = 2 << (L - 1)

            Dim v(LL - 1) As Byte

            v(0) = 10
            Dim vp As Integer = 0
            Dim vlen As Integer = 1

            For i As Integer = 1 To L
                MpMUL(v, vp, vlen, v, vp, vlen, v)
                vlen *= 2
                If v(0) = 0 Then
                    vp = 1
                    vlen -= 1
                Else
                    vp = 0
                End If
            Next

            If w Is Nothing Then
                ReDim w(vlen - 1)
                Array.Copy(v, vp, w, 0, vlen)
                wlen = vlen
                wp = 0
            Else
                ReDim Preserve w(wlen + vlen - 1)
                MpMUL(w, wp, wlen, v, vp, vlen, w)
                wlen = wlen + vlen
                If w(0) = 0 Then
                    wp = 1
                    wlen -= 1
                Else
                    wp = 0
                End If
            End If

            dc -= LL
        Loop Until dc <= 100

        Do Until dc = 0
            If wp = 0 Then ReDim Preserve w(wlen)
            MpSMU(w, wp, wlen, 10, w)
            If w(0) = 0 Then
                wp = 1
            Else
                wp = 0
                wlen += 1
            End If

            dc -= 1
        Loop

        Array.Copy(w, wp, w, 0, wlen)
        ReDim Preserve w(wlen - 1)
        Return w
    End Function
End Module
